'use strict';

exports.__esModule = true;
exports.UART = exports.REQUEST_TIMEOUT = exports.DEFAULT_RETURN_COMMAND_RESULT = exports.DEFAULT_BAUD_RATE = undefined;

var _events = require('events');

var _serialport = require('serialport');

var _serialport2 = _interopRequireDefault(_serialport);

var _fp = require('lodash/fp');

var _fp2 = _interopRequireDefault(_fp);

var _pMapSeries = require('p-map-series');

var _pMapSeries2 = _interopRequireDefault(_pMapSeries);

var _pify = require('pify');

var _pify2 = _interopRequireDefault(_pify);

var _debug = require('debug');

var _debug2 = _interopRequireDefault(_debug);

var _protocol = require('./protocol');

var _util = require('./util');

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

function _possibleConstructorReturn(self, call) { if (!self) { throw new ReferenceError("this hasn't been initialised - super() hasn't been called"); } return call && (typeof call === "object" || typeof call === "function") ? call : self; }

function _inherits(subClass, superClass) { if (typeof superClass !== "function" && superClass !== null) { throw new TypeError("Super expression must either be null or a function, not " + typeof superClass); } subClass.prototype = Object.create(superClass && superClass.prototype, { constructor: { value: subClass, enumerable: false, writable: true, configurable: true } }); if (superClass) Object.setPrototypeOf ? Object.setPrototypeOf(subClass, superClass) : subClass.__proto__ = superClass; }

/**
 * Factory default baud rate of Nextion is 9600
 * @private
 * @type {number}
 */
var DEFAULT_BAUD_RATE = exports.DEFAULT_BAUD_RATE = 9600;

/**
 * This tells the Nextion to return a command result even if command succeeded
 * @private
 * @type {string}
 */
var DEFAULT_RETURN_COMMAND_RESULT = exports.DEFAULT_RETURN_COMMAND_RESULT = 'always';

/**
 * Timeout in ms we should wait for response from Nextion for various requests
 * @private
 * @type {number}
 */
var REQUEST_TIMEOUT = exports.REQUEST_TIMEOUT = 1000;

/**
 * We use this to find a reasonable serial port
 * @private
 * @type {RegExp}
 */
var PORT_GUESS_REGEX = /usb|acm|^com/i;

/**
 * @ignore
 */
var txDebug = (0, _debug2.default)('nextion:UART:TX');

/**
 * @ignore
 */
var rxDebug = (0, _debug2.default)('nextion:UART:RX');

/**
 * @ignore
 */
var debug = (0, _debug2.default)('nextion:UART');

/**
 * Applies defaults to an object
 * @param {NextionOptions} obj - Defaults are applied to this object
 * @returns {NextionOptions} Options w/ defaults applied
 * @function
 * @private
 */
var applyDefaults = _fp2.default.defaults({
  returnCommandResult: DEFAULT_RETURN_COMMAND_RESULT,
  baudRate: DEFAULT_BAUD_RATE
});

/**
 * Whether or not to expect the Nextion device to return success/failure codes
 * after each serial command.  Defaults to "always".
 * @typedef {string|boolean|number} ReturnCommandResultValue
 * @private
 */

/**
 * Wraps a SerialPort or similar object; provides convenience methods for
 * interaction with a Nextion over UART.
 * @emits {error} When {@link UART#port} emits `error` or is unexpectedly
 *   disconnected, or when we receive unknown data from the device.
 * @emits {close} When {@link UART#close} successfully completes
 * @extends {EventEmitter}
 */

var UART = exports.UART = function (_EventEmitter) {
  _inherits(UART, _EventEmitter);

  /**
   * Sets some default options
   * @param {EventEmitter|Duplex} port - Serial port interface
   * @param {NextionOptions} [opts={}] - Options
   */
  function UART(port) {
    var opts = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};

    _classCallCheck(this, UART);

    var _this = _possibleConstructorReturn(this, _EventEmitter.call(this));

    if (!(0, _util.isValidPort)(port)) {
      throw new TypeError('"port" must be a Serialport-like object');
    }

    /**
     * Options
     * @type {NextionOptions}
     * @private
     */
    _this.opts = applyDefaults(opts);

    /**
     * Internal serial port object
     * @type {Serialport|Duplex|*}
     * @private
     */
    _this.port = port;

    /**
     * `true` once we've successfully began listening via {@link UART#bind}.
     * @type {boolean}
     */
    _this.ready = false;

    _this.on('data', function (data) {
      try {
        var result = (0, _protocol.read)(data);
        if (result.type === 'event') {
          rxDebug(`Event "${result.name}" (${result.codeByte}); data:`, result.data);
          _this.emit('event', result);
        } else {
          rxDebug(`Response "${result.name}" (${result.codeByte})`);
          _this.emit('response', result);
        }
      } catch (err) {
        _this.emit('error', err);
      }
    }).on('connect', function () {
      _this.ready = true;
    }).on('close', function () {
      _this.ready = false;
    });
    return _this;
  }

  /**
   * Set variable `variableName` to value `value`.
   * Boolean values `true` and `false` become `1` and `0`, respectively.
   * @param {string} variableName - Name of variable, component, system var,
   *   page, etc.
   * @param {*} [value] - Will be coerced to a string.
   * @returns {Promise<ResponseResult, Error>} Result of setting value, or
   *   {@link Error} if device is not ready.
   */


  UART.prototype.setValue = function setValue(variableName, value) {
    if (!this.ready) {
      return Promise.reject(new Error('Device not ready!'));
    }
    if (value === true) {
      value = 1;
    } else if (value === false) {
      value = 0;
    }
    return this.request(`${variableName}=${value}`);
  };

  /**
   * Wraps port's `write()` in a {@link Promise}
   * @param {Buffer} data - Data to write
   * @returns {Promise<UART, TypeError>} UART instance, or {@link TypeError} if
   *   `data` is not a {@link Buffer}.
   * @private
   */


  UART.prototype.write = function write(data) {
    var _this2 = this;

    return new Promise(function (resolve, reject) {
      if (!Buffer.isBuffer(data)) {
        return reject(new TypeError('Expected Buffer'));
      }
      txDebug('Writing:', data);
      _this2.port.write(data, function (err) {
        if (err) {
          return reject(err);
        }
        _this2.port.drain(function (err) {
          if (err) {
            return reject(err);
          }
          resolve();
        });
      });
    }).then(function () {
      txDebug('Wrote:', data || '(empty string)');
      return _this2;
    });
  };

  /**
   * Sends a raw command; does not wait for response.
   * @param {string} [command] - Raw ASCII command, or nothing at all
   * @private
   * @returns {Promise<UART>} This UART instance
   */


  UART.prototype.send = function send() {
    var _this3 = this;

    var command = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : '';

    txDebug('Sending:', command);
    return this.write(Buffer.concat([Buffer.from(command), _protocol.delimiterBuffer])).then(function () {
      txDebug('Sent:', command || '(empty string)');
      return _this3;
    });
  };

  /**
   * Makes a request to a Nextion device, expecting a response.
   * @param {string[]|string} commands - Command(s) to execute
   * @param {number} [timeout=1000] - How long to wait for response (in ms)
   * @returns {Promise<ResponseResult[]|ResponseResult,Error>} Result or array
   *   of results, or {@link Error} if device is not ready.
   */


  UART.prototype.request = function request() {
    var _this4 = this;

    var commands = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : [];
    var timeout = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : REQUEST_TIMEOUT;

    if (!this.ready) {
      return Promise.reject(new Error('Device not ready!'));
    }
    commands = [].concat(commands);
    return (0, _pMapSeries2.default)(commands, function (command) {
      txDebug('Beginning request');
      return new Promise(function (resolve, reject) {
        var handler = function handler(result) {
          rxDebug('Received', result);
          clearTimeout(t);
          return resolve(result);
        };

        _this4.once('response', handler);

        var t = setTimeout(function () {
          _this4.removeListener('response', handler);
          reject(new Error(`Timeout of ${timeout}ms exceeded`));
        }, timeout);

        _this4.send(command).catch(function (err) {
          _this4.removeListener('response', handler);
          reject(err);
        });
      });
    }).then(function (results) {
      return commands.length === 1 ? results.shift() : results;
    });
  };

  /**
   * Begins listening for data on the open serial port.
   * If an event is emitted from the Nextion device, an event with a
   * human-readable name will be emitted, along with any extra data, if
   * present.
   * Errors bubble up from the serial port object thru the UART instance.
   * @param {EventEmitter|Duplex} [port] - `Serialport`-like
   *   object to listen for data and errors on.  Defaults to {@link UART#port}.
   * @returns {Promise<UART>} This UART instance
   */


  UART.prototype.bind = function bind(port) {
    var _this5 = this;

    port = port || this.port;
    var nextDelimiterIndex = 0;
    var buf = [];
    port.on('data', function (data) {
      rxDebug('Raw data:', data);
      // this is torn from serialport's byte delimiter parser, considering
      // we aren't ensured a Serialport object.
      Array.from(data).forEach(function (byte) {
        buf.push(byte);
        if (_fp2.default.last(buf) === _protocol.delimiter[nextDelimiterIndex]) {
          nextDelimiterIndex++;
        }
        if (nextDelimiterIndex === _protocol.delimiter.length) {
          buf = Buffer.from(buf);
          rxDebug('Parsed data:', buf);
          _this5.emit('data', buf);
          buf = [];
          nextDelimiterIndex = 0;
        }
      });
    }).on('error', function (err) {
      _this5.emit('error', err);
    }).on('close', function () {
      _this5.emit('disconnect');
      _this5.port = null;
      delete _this5.port;
    }).on('disconnect', function (err) {
      _this5.emit('error', err);
    });

    debug('Configuring device...');
    return this.send('sleep=0').then(function () {
      return _this5.setReturnCommandResult(_this5.opts.returnCommandResult);
    }).then(function () {
      _this5.emit('connect');
      return _this5;
    });
  };

  /**
   * Tell the Nextion device to wait (default) or not wait for responses when
   * requests are made.
   * @param {string|number|boolean} [value='always'] - `false`, `0` or `'none'`
   *   to disable waiting for responses; `true`, `1` or `'always'` to enable
   *   waiting for responses.
   * @private
   * @returns {Promise<UART>} This UART instance
   */


  UART.prototype.setReturnCommandResult = function setReturnCommandResult() {
    var _this6 = this;

    var value = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : 'always';

    var commandValue = void 0;

    switch (value) {
      case false:
      case 0:
      // falls through
      case 'none':
        commandValue = 0;
        break;
      case 3:
      case true:
      // falls through
      case 'always':
      default:
        commandValue = 3;
    }

    return this.send(`bkcmd=${commandValue}`).then(function () {
      _this6.opts.returnCommandResult = value;
      debug(`returnCommandResult is now "${value}"`);
      return _this6;
    });
  };

  /**
   * Stops listening on the serial port, closes it, destroys the reference,
   * kills a kitten, etc.
   * Use {@link Nextion#close} instead.
   * @returns {Promise<UART>} This UART instance
   */


  UART.prototype.unbind = function unbind() {
    var _this7 = this;

    var port = this.port;
    this.port = null;
    delete this.port;
    if (_fp2.default.isFunction(port.close)) {
      return (0, _pify2.default)(port.close.bind(port))().then(function () {
        _this7.emit('disconnect');
        return _this7;
      });
    }
    this.emit('disconnect');
    return Promise.resolve(this);
  };

  /**
   * Given a {@link Serialport}-like object, create a {@link UART} wrapper.
   * Synchronous, but returns a {@link Promise} for consistency.
   * @param {Serialport|Duplex|*} serialPort - {@link Serialport}-like object
   * @param {NextionOptions} [opts={}] - Options
   * @returns {Promise<UART>} New {@link UART} instance
   */


  UART.fromSerial = function fromSerial(serialPort) {
    var opts = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};

    return Promise.resolve(new UART(serialPort, opts));
  };

  /**
   * Given a serial port name or path, or object containing one, create a
   * {@link Serialport} instance, open the serial port, then return a {@link
    * UART} instance. If no port name is present, we'll try to autodetect the
    * port.
   * @param {string|NextionOptions} [portName] - Serial port name or path, or
   *   options
   * @param {NextionOptions} [opts={}] - Options
   * @returns {Promise<UART>} New {@link UART} instance
   */


  UART.fromPort = function fromPort(portName) {
    var opts = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};

    if (_fp2.default.isObject(portName)) {
      opts = portName;
      portName = portName.port;
    }

    opts = applyDefaults(opts);

    if (portName) {
      return new Promise(function (resolve, reject) {
        debug(`Opening port "${portName}" with opts`, opts);
        var serialPort = new _serialport2.default(portName, opts, function (err) {
          if (err) {
            return reject(err);
          }
          resolve(serialPort);
        });
      }).then(function (serialPort) {
        debug(`Opened connection to port "${portName}"`);
        return UART.fromSerial(serialPort, _fp2.default.assign(opts, { portName }));
      });
    } else {
      return UART.findPort().then(function (portName) {
        if (!portName) {
          throw new Error('Could not find a serial device!');
        }
        return UART.fromPort(portName, opts);
      });
    }
  };

  /**
   * Convenience wrapper of {@link UART.fromPort} and {@link UART.fromSerial}.
   * @param {string|Duplex|Serialport} [port] - Port name or object
   * @param {NextionOptions} [opts] - Options
   * @returns {Promise<UART>} New {@link UART} instance
   */


  UART.from = function from(port) {
    var opts = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};

    return (0, _util.isValidPort)(port) ? UART.fromSerial(port, opts) : UART.fromPort(port, opts);
  };

  /**
   * Tries to find a device on a serial/COM port.
   * @returns {Promise<string|void>} - Name/path of promising serial port, if
   *   any
   * @private
   */


  UART.findPort = function findPort() {
    return (0, _pify2.default)(_serialport2.default.list)().then(_fp2.default.pipe(_fp2.default.pluck('comName'), _fp2.default.filter(function (portName) {
      return PORT_GUESS_REGEX.test(portName);
    }), _fp2.default.head));
  };

  return UART;
}(_events.EventEmitter);
//# sourceMappingURL=uart.js.map